from importlib import import_module
from inspect import getabsfile, isclass
from pathlib import Path
from pkgutil import walk_packages
from sys import path
from types import ModuleType
from typing import Dict, Iterator, Tuple, Type

from py2puml.domain.inspection import Inspection
from py2puml.export.puml import to_puml_content
from py2puml.inspection.inspectmodule import inspect_domain_definition


class Inspector:
    def __init__(self, inspection_working_directory: Path, root_domain_path: Path, root_domain_namespace: str):
        self.inspection_working_directory = inspection_working_directory
        self.root_domain_path = root_domain_path.resolve()
        self.root_domain_namespace = root_domain_namespace
        self.filtered_definitions: Dict[str, bool] = {}
        self.has_src_folder: bool = (inspection_working_directory / 'src').exists()

    def inspect(self, inspection: Inspection) -> Iterator[str]:
        # adds the current working directory to the system path in the first place
        # to ease module resolution when py2puml imports the module to inspect
        path.insert(0, str(self.inspection_working_directory))
        if self.has_src_folder:
            path.insert(0, str(self.inspection_working_directory / 'src'))

        if self.root_domain_path.is_dir():
            self._inspect_package(self.root_domain_path, self.root_domain_namespace, inspection)
        else:
            self._inspect_module(self.root_domain_namespace, inspection)

        yield from to_puml_content(
            self.root_domain_namespace if self.root_domain_namespace else self.root_domain_path.name,
            inspection.items_by_fqn.values(),
            inspection.relations,
        )

    def _inspect_package(self, domain_path: Path, root_domain_namespace: str, inspection: Inspection):
        # inspects the packages module and inner packages

        # inspects the package's __init__ module
        if domain_path.is_dir():
            self._inspect_module(root_domain_namespace, inspection)

        root_domain_namespace_prefix = f'{root_domain_namespace}.' if root_domain_namespace else ''
        for _, module_name, _ in walk_packages([str(domain_path)], root_domain_namespace_prefix):
            self._inspect_module(module_name, inspection)

    def _inspect_module(self, module_name: str, inspection: Inspection):
        module_to_inspect = import_module(module_name)
        for _, definition in self._filter_module_definitions(module_to_inspect):
            inspect_domain_definition(
                definition, self.root_domain_namespace, inspection.items_by_fqn, inspection.relations
            )

    def _filter_module_definitions(self, module_to_inspect: ModuleType) -> Iterator[Tuple[str, Type]]:
        for definition in vars(module_to_inspect).values():
            if isclass(definition):
                try:
                    definition_file = getabsfile(definition)
                # skips builtins
                except TypeError:
                    pass
                else:
                    # this fqn can start with "__main__" instead of the real fqn
                    definition_fqn = f'{definition.__module__}.{definition.__qualname__}'
                    yield_definition = self.filtered_definitions.get(definition_fqn)
                    if yield_definition is None:
                        try:
                            relative_path = Path(definition_file).relative_to(self.root_domain_path)
                        except ValueError:
                            pass
                            self.filtered_definitions[definition_fqn] = False
                        else:
                            definition_namespace = (
                                tuple(self.root_domain_namespace.split('.'))
                                if self.root_domain_namespace
                                else tuple(relative_path.parts)
                            )
                            definition_static_fqn = '.'.join((*definition_namespace, definition.__qualname__))

                            # filters on the specified namespace if any, or on the source filepath of the definition
                            if self.root_domain_namespace:
                                yield_definition = definition_static_fqn.startswith(self.root_domain_namespace)
                            else:
                                yield_definition = Path(definition_file).is_relative_to(self.root_domain_path)

                            self.filtered_definitions[definition_fqn] = yield_definition

                            if yield_definition:
                                yield definition_static_fqn, definition



from pathlib import Path
from typing import Iterator

from py2puml.domain.inspection import Inspection
from py2puml.inspector import Inspector


def py2puml(domain_path: str, domain_module: str, inspection_working_directory: Path = None) -> Iterator[str]:
    # uses the current working directory by default:
    if inspection_working_directory is None:
        inspection_working_directory = Path.cwd().resolve()

    yield from Inspector(inspection_working_directory, Path(domain_path), domain_module).inspect(Inspection({}, []))



from typing import Dict, List, NamedTuple

from py2puml.domain.umlitem import UmlItem
from py2puml.domain.umlrelation import UmlRelation


class Inspection(NamedTuple):
    items_by_fqn: Dict[str, UmlItem]
    relations: List[UmlRelation]



from io import StringIO
from itertools import zip_longest
from pathlib import Path
from subprocess import PIPE, run
from sys import stderr
from warnings import warn

from py2puml.controller import InspectorController


def assert_py2puml_ios(py2puml_contents: StringIO, expected_py2puml_contents: StringIO, source: str):
    """
    Compares actual against expected IO contents, typically the plantuml contents generated by a py2puml call against file contents.

    Args:
        py2puml_contents: contents streamed typically from a call to py2puml inspection
        expected_py2puml_contents: expected contents streamed typically from an opened file
        source: an arbitrary text explaining how the py2puml_contents were created (a py2puml call)

    Raises:
        ValueError: on the first line where the contents are different. The error message includes the line index.
    """
    for line_index, (output_line, expected_output_line) in enumerate(
        zip_longest(py2puml_contents, expected_py2puml_contents, fillvalue='<end of contents>'), start=1
    ):
        output_line = output_line.strip('\n')
        expected_output_line = expected_output_line.strip('\n')
        # print(f"{line_index=}", f"{output_line=}", f"{expected_output_line=}")
        if output_line != expected_output_line:
            raise ValueError(
                f"Line {line_index} is '{output_line}' but '{expected_output_line}' was expected (source: '{source}')"
            )


def assert_py2puml_command_args(
    command_args: str,
    expected_output_file_path: Path,
    current_working_directory: str = None,
    *,
    overwrite_expected_output: bool = False,
):
    """
    Runs a py2puml command in a subprocess and compares its output with the contents of the given output file.

    Args:
        command_args: the command you want to run, without the leading py2puml part
        expected_output_file_path: the path of the expected contents
        current_working_directory: the working directory where this command should be run. Leave to None if the current working directory is fine.
        overwrite_expected_output: set to true to update the file containing the expected contents

    Raises:
        ValueError: on the first line where the contents are different. The error message includes the line index.
    """

    if command_args.startswith('py2puml ') or command_args.startswith('python'):
        raise ValueError(
            f"The command should contain only the arguments without reference to the py2puml executable, got '{command_args}'"
        )

    # runs the py2puml command
    command_args_parts = command_args.split(' ')
    cli_process = run(
        ['py2puml', *command_args_parts],
        stdout=PIPE,
        stderr=PIPE,
        text=True,
        check=False,
        cwd=current_working_directory,
    )
    py2puml_command = f'py2puml {command_args}'
    if process_stderr := cli_process.stderr:
        stderr.writelines(process_stderr.splitlines(keepends=True))
        raise RuntimeError(f"An error occurred when running '{py2puml_command}'")
    parsed_args = InspectorController()._parse_args(command_args_parts)

    # verifies the generated contents, either in the written file or in the captured stdout
    expected_output_file_open_mode = 'w' if overwrite_expected_output else 'r'
    overwrite_warning = f"assert_py2puml_command_args() is used to overwrite file '{expected_output_file_path}'"
    if parsed_args.output_file:
        with (
            open(parsed_args.output_file, encoding='utf-8') as output_file,
            open(expected_output_file_path, expected_output_file_open_mode, encoding='utf-8') as expected_output_file,
        ):
            if overwrite_expected_output:
                expected_output_file.writelines(output_file)
                warn(overwrite_warning, UserWarning, stacklevel=1)
            else:
                assert_py2puml_ios(output_file, expected_output_file, py2puml_command)
    else:
        # removes the last return carriage added by the stdout
        output_io = StringIO(cli_process.stdout[:-1])

        with open(expected_output_file_path, expected_output_file_open_mode, encoding='utf-8') as expected_output_file:
            if overwrite_expected_output:
                expected_output_file.writelines(output_io)
                warn(overwrite_warning, UserWarning, stacklevel=1)
            else:
                assert_py2puml_ios(output_io, expected_output_file, py2puml_command)



from argparse import ArgumentParser
from contextlib import contextmanager
from itertools import zip_longest
from pathlib import Path
from sys import stdout
from typing import Generator, Protocol, Sequence, TextIO
from warnings import warn

from py2puml import __version__
from py2puml.domain.inspection import Inspection
from py2puml.inspector import Inspector


class InspectorArgs(Protocol):
    path: Path = None
    namespace: str = None
    output_file: Path = None


class InspectorController:
    def _parse_args(self, args: Sequence[str] = None) -> InspectorArgs:
        argparser = ArgumentParser(description='Generate PlantUML class diagrams to document your Python application.')

        argparser.add_argument('-v', '--version', action='version', version=f'py2puml {__version__}')

        # deprecated positional arguments
        argparser.add_argument(
            'deprecated_path',
            nargs='?',
            type=Path,
            metavar='path',
            help='the filepath to the domain to inspect - deprecated: use --path path/to/module... instead',
            default=None,
        )
        argparser.add_argument(
            'deprecated_namespace',
            nargs='?',
            metavar='namespace',
            help='the namespace of the domain to inspect - deprecated: use --namespace package.to.module instead',
            default=None,
        )

        # optional arguments
        argparser.add_argument(
            '-p',
            '--path',
            metavar='path',
            type=Path,
            help='the filepath to the domain to inspect. It can be a package (a folder name) or a module (a .py file).',
            default=None,
        )
        argparser.add_argument(
            '-n',
            '--namespace',
            metavar='namespace',
            help="the namespace of the domain to inspect, in case the command is not launched from the domain's root",
            default=None,
        )
        argparser.add_argument(
            '-o',
            '--output-file',
            metavar='output_file',
            dest='output_file',
            type=Path,
            help='outputs the PlantUML contents in a file if defined, or in the standard output otherwise',
            default=None,
        )

        args = argparser.parse_args(args)

        # adapts the old cli arguments if any
        if args.deprecated_path or args.deprecated_namespace:
            deprecation_message_parts = [
                'Deprecation warning: specify the path to inspect and its corresponding namespace with specific flags',
                f'Use `py2puml --path {args.deprecated_path} --namespace {args.deprecated_namespace}` instead',
                'Type py2puml --help for more details about the available flags.',
            ]
            args.path = args.deprecated_path
            args.namespace = args.deprecated_namespace

            warn('. '.join(deprecation_message_parts), DeprecationWarning, stacklevel=1)

        # remove deprecated arguments from the returned namespace
        delattr(args, 'deprecated_path')
        delattr(args, 'deprecated_namespace')

        return args

    @contextmanager
    def _open_output(self, output_file: Path = None) -> Generator[TextIO, None, None]:
        if output_file is None:
            yield stdout
        else:
            with open(output_file, 'w', encoding='utf-8') as plantuml_file:
                yield plantuml_file

    def inspect(self, args: Sequence[str] = None):
        parsed_args = self._parse_args(args)

        # inspects the current working directory if no path is specified
        current_working_directory = Path.cwd().resolve()
        root_domain_path = parsed_args.path or current_working_directory

        # eases the default case where the domain namespace is the relative namespace
        # between the inspected path and the current working directory
        inspection_working_directory = current_working_directory
        if parsed_args.namespace is None:
            namespace_relative_path = root_domain_path.resolve().relative_to(current_working_directory)
            if root_domain_path.is_dir():
                namespace_parts = namespace_relative_path.parts
            else:
                namespace_parts = [*namespace_relative_path.parent.parts, namespace_relative_path.stem]
            root_domain_namespace = '.'.join(namespace_parts)
        else:
            root_domain_namespace = parsed_args.namespace
            cwd_diff_path = self._check_domain_path_and_namespace_consistency(root_domain_path, root_domain_namespace)
            inspection_working_directory = inspection_working_directory.joinpath(cwd_diff_path).resolve()

        with self._open_output(parsed_args.output_file) as output_io:
            for plantuml_line in Inspector(
                inspection_working_directory, root_domain_path, root_domain_namespace
            ).inspect(Inspection({}, [])):
                output_io.write(plantuml_line)

    def _check_domain_path_and_namespace_consistency(self, root_domain_path: Path, root_domain_namespace: str) -> Path:
        """
        Checks that the path to the domain to inspect matches its namespace.

        Raises:
            ValueError: if there is an inconsistency between the path and the namespace

        Returns:
            the path difference from the current working directory in which the inspection should start
        """
        namespace_parts = root_domain_namespace.split('.') if root_domain_namespace else ()
        if root_domain_path.is_file():
            root_domain_path_parts = [*root_domain_path.parent.parts, root_domain_path.stem]
        else:
            root_domain_path_parts = root_domain_path.parts

        relative_cwd_change = []
        for from_end_index, (path_part, namespace_part) in enumerate(
            zip_longest(reversed(root_domain_path_parts), reversed(namespace_parts), fillvalue=None)
        ):
            # moves the cwd to the parent to reach the root of the domain namespace
            if path_part is None:
                relative_cwd_change.insert(0, '..')
            # moves the cwd deeper to reach the root of the domain namespace
            elif namespace_part is None:
                relative_cwd_change.insert(0, path_part)
            # raises an error if there is a mismatch between the path part and the namespace part
            elif path_part != namespace_part:
                subpath = '/'.join(root_domain_path.parts[:-from_end_index])
                path_and_namespace_error = f"the namespace part '{namespace_part}' of namespace '{root_domain_namespace}' does not match subpath '{subpath}' of '{str(root_domain_path)}'"
                raise ValueError(path_and_namespace_error)

        return Path(*relative_cwd_change)
